import pytest
from rraft import Progress


@pytest.mark.benchmark(group="progress", warmup=True)
def test_progress_creation(benchmark):
    def progress_creation():
        Progress(9, 10)

    benchmark(progress_creation)



import warnings

import pytest
from harness.src.network import Network

from rraft import (
    MessageType,
    ProgressState,
    Snapshot,
    default_logger,
    INVALID_INDEX,
    RequestSnapshotDroppedError,
)

from harness.utils import (
    new_snapshot,
    new_storage,
    new_test_raft,
    new_message,
    new_test_raft_with_prevote,
)


def testing_snap() -> Snapshot:
    warnings.filterwarnings("ignore")
    return new_snapshot(11, 11, [1, 2])


def test_sending_snapshot_set_pending_snapshot():
    l = default_logger()
    storage = new_storage()
    sm = new_test_raft(1, [1, 2], 10, 1, storage, l)
    sm.raft.restore(testing_snap())
    sm.persist()

    sm.raft.become_candidate()
    sm.raft.become_leader()

    # force set the next of node 1, so that
    # node 1 needs a snapshot
    assert sm.raft.prs().get(2).get_next_idx() == sm.raft_log.first_index()

    m = new_message(2, 1, MessageType.MsgAppendResponse, 0)
    voter_2 = sm.raft.prs().get(2)
    m.set_index(voter_2.get_next_idx() - 1)
    m.set_reject(True)

    sm.step(m)
    assert sm.raft.prs().get(2).get_pending_snapshot() == 11


def test_pending_snapshot_pause_replication():
    l = default_logger()
    storage = new_storage()
    sm = new_test_raft(1, [1, 2], 10, 1, storage, l)
    sm.raft.restore(testing_snap())
    sm.persist()

    sm.raft.become_candidate()
    sm.raft.become_leader()
    sm.raft.prs().get(2).become_snapshot(11)

    sm.step(new_message(2, 1, MessageType.MsgPropose, 1))
    msgs = sm.read_messages()
    assert not msgs


def test_snapshot_failure():
    l = default_logger()
    storage = new_storage()
    sm = new_test_raft(1, [1, 2], 10, 1, storage, l)
    sm.raft.restore(testing_snap())
    sm.persist()

    sm.raft.become_candidate()
    sm.raft.become_leader()

    sm.raft.prs().get(2).set_next_idx(1)
    sm.raft.prs().get(2).become_snapshot(11)

    m = new_message(2, 1, MessageType.MsgSnapStatus, 0)
    m.set_reject(True)
    sm.step(m)
    voter_2 = sm.raft.prs().get(2)

    assert voter_2.get_pending_snapshot() == 0
    assert voter_2.get_next_idx() == 1
    assert voter_2.get_paused()


def test_snapshot_succeed():
    l = default_logger()
    storage = new_storage()
    sm = new_test_raft(1, [1, 2], 10, 1, storage, l)
    sm.raft.restore(testing_snap())
    sm.persist()

    sm.raft.become_candidate()
    sm.raft.become_leader()

    sm.raft.prs().get(2).set_next_idx(1)
    sm.raft.prs().get(2).become_snapshot(11)

    m = new_message(2, 1, MessageType.MsgSnapStatus, 0)
    m.set_reject(False)
    sm.step(m)
    voter_2 = sm.raft.prs().get(2)
    assert voter_2.get_pending_snapshot() == 0
    assert voter_2.get_next_idx() == 12
    assert voter_2.get_paused()


def test_snapshot_abort():
    l = default_logger()
    storage = new_storage()
    sm = new_test_raft(1, [1, 2], 10, 1, storage, l)
    sm.raft.restore(testing_snap())
    sm.persist()

    sm.raft.become_candidate()
    sm.raft.become_leader()

    sm.raft.prs().get(2).set_next_idx(1)
    sm.raft.prs().get(2).become_snapshot(11)

    m = new_message(2, 1, MessageType.MsgSnapStatus, 0)
    m.set_reject(False)
    sm.step(m)

    assert sm.raft.prs().get(2).get_pending_snapshot() == 0
    assert sm.raft.prs().get(2).get_next_idx() == 12


# Initialized storage should be at term 1 instead of 0. Otherwise the case will fail.
def test_snapshot_with_min_term():
    l = default_logger()

    def do_test(pre_vote: bool):
        s1 = new_storage()
        s1.wl().apply_snapshot(new_snapshot(1, 1, [1, 2]))

        n1 = new_test_raft_with_prevote(1, [1, 2], 10, 1, s1, pre_vote, l)

        s2 = new_storage()
        n2 = new_test_raft_with_prevote(2, [], 10, 1, s2, pre_vote, l)

        nt = Network.new([n1, n2], l)
        m = new_message(1, 1, MessageType.MsgHup, 0)
        nt.send([m])

        # 1 will be elected as leader, and then send a snapshot and an empty entry to 2.
        assert nt.peers.get(2).raft_log.first_index() == 2
        assert nt.peers.get(2).raft_log.last_index() == 2

    do_test(True)
    do_test(False)


def test_request_snapshot():
    l = default_logger()
    storage = new_storage()
    sm = new_test_raft(1, [1, 2], 10, 1, storage, l)
    sm.raft.restore(testing_snap())
    sm.persist()

    # Raft can not step request snapshot if last raft log's term mismatch current term.
    with pytest.raises(RequestSnapshotDroppedError):
        sm.raft.request_snapshot()

    sm.raft.become_candidate()
    sm.raft.become_leader()

    # Raft can not step request snapshot if itself is a leader.
    with pytest.raises(RequestSnapshotDroppedError):
        sm.raft.request_snapshot()

    m = new_message(2, 1, MessageType.MsgAppendResponse, 0)
    m.set_index(11)
    sm.step(m)

    assert sm.raft.prs().get(2).get_state() == ProgressState.Replicate

    request_snapshot_idx = sm.raft_log.get_committed()

    m = new_message(2, 1, MessageType.MsgAppendResponse, 0)
    m.set_index(11)
    m.set_reject(True)
    m.set_reject_hint(INVALID_INDEX)
    m.set_request_snapshot(request_snapshot_idx)

    # Ignore out of order request snapshot messages.
    out_of_order = m.clone()
    out_of_order.set_index(9)
    sm.step(out_of_order)
    assert sm.raft.prs().get(2).get_state() == ProgressState.Replicate

    # Request snapshot.
    sm.step(m)
    assert sm.raft.prs().get(2).get_state() == ProgressState.Snapshot
    assert sm.raft.prs().get(2).get_pending_snapshot() == 11
    assert sm.raft.prs().get(2).get_next_idx() == 12
    assert sm.raft.prs().get(2).is_paused()
    snap = sm.raft.get_msgs()[-1]
    assert snap.get_msg_type() == MessageType.MsgSnapshot
    assert snap.get_snapshot().get_metadata().get_index() == request_snapshot_idx

    # Append/heartbeats does not set the state from snapshot to probe.
    m = new_message(2, 1, MessageType.MsgAppendResponse, 0)
    m.set_index(11)
    sm.step(m)
    assert sm.raft.prs().get(2).get_state() == ProgressState.Snapshot
    assert sm.raft.prs().get(2).get_pending_snapshot() == 11
    assert sm.raft.prs().get(2).get_next_idx() == 12
    assert sm.raft.prs().get(2).is_paused()

    # However snapshot status report does set the stat to probe.
    m = new_message(2, 1, MessageType.MsgSnapStatus, 0)
    sm.step(m)
    assert sm.raft.prs().get(2).get_state() == ProgressState.Probe
    assert sm.raft.prs().get(2).get_pending_snapshot() == 0
    assert sm.raft.prs().get(2).get_next_idx() == 12
    assert sm.raft.prs().get(2).is_paused()



from harness.utils import new_message, new_storage, new_test_raft
from rraft import MessageType, default_logger


# test_msg_app_flow_control_full ensures:
# 1. msgApp can fill the sending window until full
# 2. when the window is full, no more msgApp can be sent.
def test_msg_app_flow_control_full():
    l = default_logger()
    storage = new_storage()
    r = new_test_raft(
        id=1,
        peers=[1, 2],
        election=5,
        heartbeat=1,
        storage=storage,
        logger=l,
    )

    r.raft.become_candidate()
    r.raft.become_leader()

    # force the progress to be in replicate state
    r.raft.prs().get(2).become_replicate()

    # fill in the inflights window
    for i in range(0, r.raft.get_max_inflight()):
        r.raft.step(new_message(1, 1, MessageType.MsgPropose, 1))
        ms = r.read_messages()
        assert len(ms) == 1, f"#{i}: ms count = {ms.len()}, want 1"

    # ensure 1
    assert r.raft.prs().get(2).get_ins().full()

    # ensure 2
    for i in range(0, 10):
        r.step(new_message(1, 1, MessageType.MsgPropose, 1))
        ms = r.read_messages()
        assert len(ms) == 0, f"#{i}: ms count = {len(ms)}, want 0"


# test_msg_app_flow_control_move_forward ensures msgAppResp can move
# forward the sending window correctly:
# 1. valid msgAppResp.index moves the windows to pass all smaller or equal index.
# 2. out-of-dated msgAppResp has no effect on the sliding window.
def test_msg_app_flow_control_move_forward():
    l = default_logger()
    storage = new_storage()
    r = new_test_raft(
        id=1,
        peers=[1, 2],
        election=5,
        heartbeat=1,
        storage=storage,
        logger=l,
    )

    r.raft.become_candidate()
    r.raft.become_leader()

    # force the progress to be in replicate state
    r.raft.prs().get(2).become_replicate()

    # fill in the inflights window
    for i in range(0, r.raft.get_max_inflight()):
        r.raft.step(new_message(1, 1, MessageType.MsgPropose, 1))
        r.read_messages()

    # 1 is noop, 2 is the first proposal we just sent.
    # so we start with 2.
    for tt in range(2, r.raft.get_max_inflight()):
        m = new_message(2, 1, MessageType.MsgAppendResponse, 0)
        m.set_index(tt)
        r.step(m)
        r.read_messages()

        # fill in the inflights window again
        r.step(new_message(1, 1, MessageType.MsgPropose, 1))
        ms = r.read_messages()
        assert len(ms) == 1, f"#{tt}: ms count = {len(ms)}, want 1"

        # ensure 1
        assert r.raft.prs().get(2).get_ins().full()

        # ensure 2
        for i in range(0, tt):
            m = new_message(2, 1, MessageType.MsgAppendResponse, 0)
            m.set_index(i)
            r.step(m)
            assert (
                r.raft.prs().get(2).get_ins().full()
            ), f"#{tt}: inflights.full = {r.raft.prs().get(2).get_ins().full()}, want true"


# test_msg_app_flow_control_recv_heartbeat ensures a heartbeat response
# frees one slot if the window is full.
def test_msg_app_flow_control_recv_heartbeat():
    l = default_logger()
    storage = new_storage()
    r = new_test_raft(
        id=1,
        peers=[1, 2],
        election=5,
        heartbeat=1,
        storage=storage,
        logger=l,
    )

    r.raft.become_candidate()
    r.raft.become_leader()

    # force the progress to be in replicate state
    r.raft.prs().get(2).become_replicate()

    # fill in the inflights window
    for _ in range(0, r.raft.get_max_inflight()):
        r.step(new_message(1, 1, MessageType.MsgPropose, 1))
        r.read_messages()

    for tt in range(1, 5):
        assert (
            r.raft.prs().get(2).get_ins().full()
        ), f"#{tt}: inflights.full = False, want True"

        # recv tt MsgHeartbeatResp and expect one free slot
        for i in range(0, tt):
            r.step(new_message(2, 1, MessageType.MsgHeartbeatResponse, 0))
            r.read_messages()
            assert (
                not r.raft.prs().get(2).get_ins().full()
            ), f"#{tt}.{i}: inflights.full = True, want False"

        # one slot
        r.step(new_message(1, 1, MessageType.MsgPropose, 1))
        ms = r.read_messages()
        assert len(ms) == 1, f"#{tt}: free slot = 0, want 1"

        # and just one slot
        for i in range(0, 10):
            r.step(new_message(1, 1, MessageType.MsgPropose, 1))
            msl = r.read_messages()
            assert len(msl) == 0, f"#{tt}.{i}, ms1 should be empty."

        # clear all pending messages
        r.step(new_message(2, 1, MessageType.MsgHeartbeatResponse, 0))
        r.read_messages()


def test_msg_app_flow_control_with_freeing_resources():
    l = default_logger()
    s = new_storage()
    r = new_test_raft(1, [1, 2, 3], 5, 1, s, l)

    r.raft.become_candidate()
    r.raft.become_leader()

    for pr in r.raft.prs().collect():
        assert not pr.progress().get_ins().buffer_is_allocated()

    for i in range(1, 3 + 1):
        # Force the progress to be in replicate state.
        r.raft.prs().get(i).become_replicate()

    r.raft.step(new_message(1, 1, MessageType.MsgPropose, 1))

    for pr in r.raft.prs().collect():
        if pr.id() != 1:
            assert pr.progress().get_ins().buffer_is_allocated()
            assert pr.progress().get_ins().count() == 1

    # 1: cap=0/start=0/count=0/buffer=[]
    # 2: cap=256/start=0/count=1/buffer=[2]
    # 3: cap=256/start=0/count=1/buffer=[2]

    resp = new_message(2, 1, MessageType.MsgAppendResponse, 0)
    resp.set_index(r.raft_log.last_index())
    r.step(resp)

    assert r.raft.prs().get(2).get_ins().count() == 0

    # 1: cap=0/start=0/count=0/buffer=[]
    # 2: cap=256/start=1/count=0/buffer=[2]
    # 3: cap=256/start=0/count=1/buffer=[2]

    r.step(new_message(1, 1, MessageType.MsgPropose, 1))

    assert r.raft.prs().get(2).get_ins().count(), 1
    assert r.raft.prs().get(3).get_ins().count(), 2

    # 1: cap=0/start=0/count=0/buffer=[]
    # 2: cap=256/start=1/count=1/buffer=[2,3]
    # 3: cap=256/start=0/count=2/buffer=[2,3]

    resp = new_message(2, 1, MessageType.MsgAppendResponse, 0)
    resp.set_index(r.raft_log.last_index())
    r.step(resp)

    assert r.raft.prs().get(2).get_ins().count() == 0
    assert r.raft.prs().get(3).get_ins().count() == 2
    assert r.raft.inflight_buffers_size() == 4096

    # 1: cap=0/start=0/count=0/buffer=[]
    # 2: cap=256/start=2/count=0/buffer=[2,3]
    # 3: cap=256/start=0/count=2/buffer=[2,3]

    r.raft.maybe_free_inflight_buffers()

    assert not r.raft.prs().get(2).get_ins().buffer_is_allocated()
    assert r.raft.prs().get(2).get_ins().count() == 0
    assert r.raft.inflight_buffers_size() == 2048

    # 1: cap=0/start=0/count=0/buffer=[]
    # 2: cap=0/start=0/count=0/buffer=[]
    # 3: cap=256/start=0/count=2/buffer=[2,3]


def test_disable_progress():
    l = default_logger()
    s = new_storage()
    r = new_test_raft(1, [1, 2], 5, 1, s, l)
    r.raft.become_candidate()
    r.raft.become_leader()

    r.raft.prs().get(2).become_replicate()

    # Disable the progress 2. Internal `free`s shouldn't fail.
    r.raft.adjust_max_inflight_msgs(2, 0)
    r.step(new_message(2, 1, MessageType.MsgHeartbeatResponse, 0))

    assert r.raft.prs().get(2).get_ins().full()
    assert not r.raft.prs().get(2).get_ins().count()

    # Progress 2 is disabled.
    msgs = r.read_messages()
    assert not msgs

    # After the progress gets enabled and a heartbeat response is received,
    # its leader can continue to append entries to it.
    r.raft.adjust_max_inflight_msgs(2, 10)
    r.step(new_message(2, 1, MessageType.MsgHeartbeatResponse, 0))

    msgs = r.read_messages()
    assert len(msgs) == 1
    assert msgs[0].get_msg_type() == MessageType.MsgAppend



from typing import Dict, List, Optional, cast
from harness.src.interface import Interface
from rraft import (
    Config,
    ConfState,
    ConfigRef,
    Logger,
    LoggerRef,
    MemStorage,
    Message,
    MessageRef,
    MessageType,
    NO_LIMIT,
    InMemoryRaft,
)
import random


class Connection:
    """
    A connection from one node to another.

    Used in by `Network` for determining drop rates on messages.
    """
    def __hash__(self) -> int:
        return hash((self.from_, self.to))

    def __eq__(self, other) -> bool:
        return self.from_ == other.from_ and self.to == other.to

    def __init__(self, from_: int, to: int) -> None:
        self.from_ = from_
        self.to = to


class Network:
    """
    A simulated network for testing.

    You can use this to create a test network of Raft nodes.

    *Please note:* no actual network calls are made.
    """
    def __init__(
        self,
        peers: Dict[int, Interface],
        storage: Dict[int, MemStorage],
        dropm: Dict[Connection, float],
        ignorem: Dict[MessageType, bool],
    ) -> None:
        self.peers = peers
        self.storage = storage
        self.dropm = dropm
        self.ignorem = ignorem

    @staticmethod
    def default_config() -> Config:
        """
        Get a base config. Calling `Network::new` will initialize peers with this config.
        """
        cfg = Config.default()
        cfg.set_election_tick(10)
        cfg.set_heartbeat_tick(1)
        cfg.set_max_size_per_msg(NO_LIMIT)
        cfg.set_max_inflight_msgs(256)
        return cfg

    @staticmethod
    def new(peers: List[Optional[Interface]], l: Logger | LoggerRef) -> "Network":
        """
        Initializes a network from `peers`.

        Nodes will receive their ID based on their index in the vector, starting with 1.

        A `None` node will be replaced with a new Raft node, and its configuration will
        be `peers`.
        """
        return Network.new_with_config(peers, Network.default_config(), l)

    @staticmethod
    def new_with_config(
        peers: List[Optional[Interface]],
        config: Config | ConfigRef,
        l: Logger | LoggerRef,
    ) -> "Network":
        """
        Initialize a network from `peers` with explicitly specified `config`.
        """
        nstorage = {}
        npeers = {}
        peer_addrs = list(range(1, len(peers) + 1))

        for p, id in zip(peers, peer_addrs):
            if p is None:
                store = MemStorage.new_with_conf_state(
                    ConfState(peer_addrs, [])
                )
                nstorage[id] = store.clone()
                cfg = config.clone()
                cfg.set_id(id)
                npeers[id] = Interface(InMemoryRaft(cfg, store, l))
            else:
                if p.raft:
                    if raft := p.raft:
                        assert (
                            raft.get_id() == id
                        ), f"peer {p.raft.get_id()} in peers has a wrong position"

                        store = raft.get_raft_log().get_store().clone()
                        nstorage[id] = store

                npeers[id] = p

        return Network(npeers, nstorage, {}, {})

    def ignore(self, t: MessageType) -> None:
        """
        Ignore a given `MessageType`.
        """
        self.ignorem[t] = True

    def filter_(
        self, msgs: List[Message] | List[MessageRef]
    ) -> List[Message]:
        """
        Filter out messages that should be dropped according to rules set by `ignore` or `drop`.
        """
        def should_be_filtered(m: Message | MessageRef):
            if self.ignorem.get(m.get_msg_type()):
                return False

            # hups never go over the network, so don't drop them but panic
            assert m.get_msg_type() != MessageType.MsgHup, "unexpected msgHup"

            perc = self.dropm.get(Connection(m.get_from(), m.get_to()), 0.0)

            return random.random() >= perc

        return list(filter(should_be_filtered, msgs))

    def read_messages(self) -> List[Message]:
        """
        # Read out all messages generated by peers in the `Network`.
        #
        # Note: messages are not filtered by any configured filters.
        """
        msgs = []
        for _id, p in self.peers.items():
            msgs.extend(p.read_messages())
        return msgs

    def send(self, msgs: List[Message] | List[MessageRef]) -> None:
        """
        # Instruct the cluster to `step` through the given messages.
        #
        # NOTE: the given `msgs` won't be filtered by its filters.
        """
        while msgs:
            new_msgs: List[Message] = []

            for m in msgs:
                p = cast(Interface, self.peers.get(m.get_to()))
                try:
                    p.step(m)
                except Exception:
                    continue
                finally:
                    # The unstable data should be persisted before sending msg.
                    p.persist()
                    resp = p.read_messages()
                    new_msgs.extend(self.filter_(resp))

            msgs = []
            msgs.extend(new_msgs)

    def filter_and_send(self, msgs: List[Message] | List[MessageRef]) -> None:
        """
        Filter `msgs` and then instruct the cluster to `step` through the given messages.
        """
        self.send(self.filter_(msgs))

    def dispatch(self, messages: List[Message] | List[MessageRef]) -> None:
        """
        # Dispatches the given messages to the appropriate peers.
        #
        # Unlike `send` this does not gather and send any responses. It also does not ignore errors.]
        """
        for message in self.filter_(messages):
            to = message.get_to()
            peer = self.peers[to]
            peer.step(message)

    def drop(self, from_: int, to: int, perc: float) -> None:
        """
        # Ignore messages from `from` to `to` at `perc` percent chance.
        #
        # `perc` set to `1f64` is a 100% chance, `0f64` is a 0% chance.
        """
        self.dropm[Connection(from_, to)] = perc

    def cut(self, one: int, other: int) -> None:
        """
        Cut the communication between the two given nodes.
        """
        self.drop(one, other, 1.0)
        self.drop(other, one, 1.0)

    def isolate(self, id: int) -> None:
        """
        Isolate the given raft to and from all other raft in the cluster.
        """
        for i in range(0, len(self.peers)):
            nid = i + 1

            if nid != id:
                self.drop(id, nid, 1.0)
                self.drop(nid, id, 1.0)

    def recover(self) -> None:
        """
        Recover the cluster conditions applied with `drop` and `ignore`.
        """
        self.dropm = {}
        self.ignorem = {}
